(* =========== Basic tactics :
-> intro - introduction rule for Pi;
-> apply - elimination rule for Pi;
-> assumption, exact - match conclusion with an hypothesis. 
*)

(* =========== Tactics for EQUALITY and REWRITING :
-> rewrite - rewrites an equality;
-> rewrite <- - reverse rewrite of an equality;
-> reflexivity - reflexivity property for equality;
-> symmetry - symmetry property for equality;
-> transitivity - transitivity property for equality.
*)

(* =========== Tactics for EVALUATION and CONVERTIBILITY :
-> simpl, red, cbv, lazy - performs evaluation;
-> pattern - performs a beta-expansion on the goal;
-> change - replaces the goal by a convertible one.
*)

(* =========== Tactics for INDUCTION :
-> elim - to apply the corresponding induction principle;
-> induction - performs induction on an identifier;
-> destruct - case analysis;
-> discriminate - discriminates objects built from different constructors;
-> injection - constructors of inductive types are injections; 
-> inversion - given an inductive type instance, find all the necessary condition 
               that must hold on the arguments of its constructors.
*)

(* =========== LIBRARIES 
A large base of definitions and facts found in the Coq Standard Library.
Often used libraries: 
-> Arith - unary integers;
-> ZArith - binary integers;
-> List - polymorphic lists;

Useful commands for finding theorems acting on a given identifier:
-> Search
-> SearchAbout
-> SearchPattern
*)

(* =========== AUTOMATISATION
For some specific domains, Coq is able to support some degree of automatisation:
-> auto - automatically applies theorems from a database;
-> tauto, intuition - decision procedures for specific classes of goals (e.g. propositional logic);
-> omega, ring - specialized tactics for numerical properties.
*)

(* =========== USEFUL tactics and commands...
Tactics:
-> clear - removes an hypothesis from the environment;
-> generalize - re-introduce an hypothesis into the goal;
-> cut, assert - proves the goal through an intermediate result;
-> pattern - performs eta-expansion on the goal.

Commands:
-> Admitted - aborts the current proof (property is assumed);
-> Set Implicit Arguments - makes possible to omit some arguments (when inferable by the system);
-> Open Scope - opens a syntax notation scope (constants, operators, etc.)

See the Reference Manual...
*)



(* ================================================================== *)
(* =============== Reasoning about naturals and lists =============== *)
(* ================================================================== *)

Require Import List.

Set Implicit Arguments.

Print List.
Print app.


Theorem app_nil : forall (A:Type) (l:list A), app l nil = l.
intros A l.
elim l.
reflexivity.
intros a l0 IH. simpl. rewrite IH. reflexivity.
Qed.

(* de outro modo *)
Theorem app_nil' : forall (A:Type) (l:list A), app l nil = l.
intros A l. 
pattern l. (* pattern performs a beta-expansion *)
apply list_ind.
reflexivity.
intros a l0 IH; simpl; rewrite IH; reflexivity.
Abort.

Print nat.
Print length.


(* 
 Exercise: prove that the length of the append of two lists is equal 
           to the addition of the length of each list 
*)





(*
 Exercise: use the lemma below to prove the following theorem

      Theorem rev_rev : forall A:Type, forall l:list A, rev (rev l) = l.
*)

Lemma rev_aux : forall (A:Type) (l:list A) a, rev(l ++ (a::nil)) = a::(rev l).
induction l. auto.
intros. simpl.
rewrite IHl.
auto.
Qed.





(* 
 Exercise: prove the following theorem

 Theorem rev_app : forall A:Type, forall l1 l2:list A, rev (app l1 l2) = app (rev l2) (rev l1).

 Hint: you will need a lemma for associativity of ++. 
       Use SearchPattern to find out the name of the lemma.
*)






(* Inductive predicates *)


(* A property (predicate) on natural numbers *)
Inductive Even : nat -> Prop :=
| Even_base : Even 0
| Even_step : forall n, Even n -> Even (S (S n)).


Lemma not_1_even : ~(Even 1).
red.
intros.
inversion H.
Qed.


Print plus.
Print mult.


Lemma double_even : forall n:nat, Even (2*n).
intros.
simpl.
elim n.
simpl.
apply Even_base. (* constructor *)

intros.
simpl.

cut (forall x y:nat, x+(S y) = S (x+y)).

intros.
rewrite H0.
constructor. (* apply Even_step. *)
assumption.

clear H.
intros.
induction x.
  simpl.
  reflexivity.

  simpl.
  rewrite IHx.
  reflexivity.
Qed.


(*
 Exercise: rewrite the prove of the previous lemma without using the cut tactic.
           Hint: use the command SearchAbout to find out a useful lemma in the database.

 Lemma double_even' : forall n:nat, Even (2*n).
*)




(*
 Exercise: prove the following proposition

 Proposition sum_even : forall n m:nat, Even n -> Even m -> Even (n+m).
*)





(*
 Exercise: define the "Odd" predicate and prove that for every n, (Even n)->(Odd (S n)).
*)





(* An inductive relation "x is the last element of list l" *)
Inductive last (A:Type) (x:A) : list A -> Prop :=
| last_base : last x (cons x nil)
| last_step : forall l y, last x l -> last x (cons y l).


Lemma last_nil : forall (A:Type) (x:A), ~(last x nil).
intros.
intro.
inversion H.
Qed.


(* The prove of  ~(last x nil)  without using the inversion tactic. *)

Lemma last_nil_aux : forall (A:Type) l (x:A), last x l -> l=nil -> False.
intros A l x H. elim H.
  intro H0. discriminate H0.
  intros. discriminate H2.
Qed.

Theorem last_nil' : forall (A:Type) (x:A), ~(last x nil).
intros. red. intros.
apply last_nil_aux with A nil x.
assumption.
reflexivity.
Qed.




(* An inductive relation "y belongs to list l" *)
Inductive InL (A:Type) (y:A) : list A -> Prop :=
| InHead : forall xs:list A, InL y (cons y xs)
| InTail : forall (x:A) (xs:list A), InL y xs -> InL y (cons x xs).



Lemma in_app1 : forall (A:Type) (l1 l2:list A) (a:A), 
                           InL a (app l1 l2) -> InL a l1 \/ InL a l2.
Proof.
induction l1.
  idtac "caso base".
  intros. right.
  simpl in H. assumption.

  idtac "passo de induo".
  intros.
  simpl in H.
  (* um exemplo de uma inverso que gera dois objectivos... *)
  inversion H.
    (* caso InHead *) left. constructor.
    (* caso InTail *) 
      elim (IHl1 _ _ H1). 
        intro H3. left. constructor. assumption.
        intro. right. assumption.
Qed.


Lemma in_app2 : forall (A:Set) (l1 l2:list A) (a:A), 
                           InL a l1 \/ InL a l2 -> InL a (app l1 l2).
Proof.
induction l1.
  (* caso base *)
  intros. simpl.
  destruct H.
    inversion_clear H.
    assumption.
  (* passo de induo *)
  intros. simpl.
  destruct H.
  inversion_clear H.
    constructor. apply InTail. apply IHl1. left. assumption.
    constructor. apply IHl1. right. assumption.
Qed.

